import requests
from oauthlib.oauth2 import (
    AccessDeniedError,
    BackendApplicationClient,
    UnauthorizedClientError,
)
from requests.auth import HTTPBasicAuth
from requests_oauthlib import OAuth2Session

from helloasso_api.exceptions import (
    ApiV5AuthenticationError,
    ApiV5ConnectionError,
    Apiv5ExceptionError,
    ApiV5Timeout,
)
from helloasso_api.utils import get_log


class OAuth2Api(object):
    """Handle Authentication logic"""
    def __init__(
        self,
        api_base: str,
        client_id: str,
        client_secret: str,
        timeout: float,
        access_token: str = None,
        refresh_token: str = None,
        oauth2_token_getter: callable = None,
        oauth2_token_setter: callable = None,
    ):
        self.api_base = api_base
        self.client_id = client_id
        self.client_secret = client_secret
        self.timeout = timeout
        self._access_token = access_token
        self._refresh_token = refresh_token
        self.oauth2_token_getter = oauth2_token_getter
        self.oauth2_token_setter = oauth2_token_setter
        self.client = BackendApplicationClient(client_id=client_id)
        self.auth = HTTPBasicAuth(client_id, client_secret)
        self.log = get_log("apiv5.oauth2")

    def _get_path(self) -> str:
        return f"https://{self.api_base}/oauth2/token"

    @staticmethod
    def _get_headers() -> dict:
        return {
            "cache-control": "no-cache",
            "content-type": "application/x-www-form-urlencoded",
        }

    @property
    def access_token(self) -> str:
        """Return an access token. If a getter has been provided at instantiation it will be used,
        else the property _access_token will be used.
        """
        if self.oauth2_token_getter:
            return (
                self.oauth2_token_getter("access_token", self.client_id)
                or self._access_token
            )
        return self._access_token

    @access_token.setter
    def access_token(self, access_token: str):
        """Set the access token. If a setter has been provided at instantiation it will be used."""
        if self.oauth2_token_setter:
            self.oauth2_token_setter("access_token", self.client_id, access_token)
        self._access_token = access_token

    @property
    def refresh_token(self) -> str:
        """Return a refresh token. If a getter has been provided at instantiation it will be used,
        else the property _refresh_token will be used.
        """
        if self.oauth2_token_getter:
            return (
                self.oauth2_token_getter("refresh_token", self.client_id)
                or self._refresh_token
            )
        return self._refresh_token

    @refresh_token.setter
    def refresh_token(self, refresh_token: str):
        """Set the refresh token. If a setter has been provided at instantiation it will be used."""
        if self.oauth2_token_setter:
            self.oauth2_token_setter("refresh_token", self.client_id, refresh_token)
        self._refresh_token = refresh_token

    @property
    def credentials(self) -> dict:
        """Return the payload dict to authenticate to the Api."""
        return {
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "refresh_token": self.refresh_token,
        }

    def get_token(self) -> None:
        """Authenticate to ApiV5 to get an access and a refresh token.
        HTTP errors are mapped to Python exceptions.
        """
        self.log.info("OAUTH2 : Get Token")
        try:
            oauth = OAuth2Session(client=self.client)
            result = oauth.fetch_token(
                token_url=self._get_path(), auth=self.auth, timeout=self.timeout
            )
            self.token_saver(result)
            self.log.info(f"Token : {self._access_token}")
        except requests.exceptions.ConnectionError:
            raise ApiV5ConnectionError(
                f"Failed to establish a new connection: Name or service not known : {self._get_path()}"
            )
        except requests.exceptions.Timeout:
            raise ApiV5Timeout(f"{self._get_path()} timeout : {str(self.timeout)} sec")
        except UnauthorizedClientError as e:
            raise ApiV5AuthenticationError(f"Authentication Error : {str(e)}")
        except Exception as e:
            raise Apiv5ExceptionError(f"Error : {str(e)}")

    def token_saver(self, request):
        """Parse dict response from oauth2 library and store access and refresh tokens."""
        self.access_token = request["access_token"]
        self.refresh_token = request["refresh_token"]

    def refresh_tokens(self):
        """Refresh connection tokens. If tokens are not presents a new token will be requested instead.
        HTTP errors are mapped to Python exceptions.
        """
        self.log.info("OAUTH2 : Refresh Token")
        try:
            if self.refresh_token is not None:
                oauth = OAuth2Session(client=self.client, token=self.access_token)
                result = oauth.refresh_token(
                    token_url=self._get_path(), timeout=self.timeout, **self.credentials
                )
                self.token_saver(result)
                self.log.info(f"OAUTH2 : Refresh Token : {self._access_token}")
            else:
                self.access_token, self.refresh_token = None, None
                self.log.warning(f"OAUTH2 : the Refresh Token is empty, reset tokens.")
        except requests.exceptions.ConnectionError:
            raise ApiV5ConnectionError(
                f"OAUTH2 : Failed to establish a new connection: Name or service not known : {self._get_path()}"
            )
        except requests.exceptions.Timeout:
            raise ApiV5Timeout(
                f"OAUTH2 : {self._get_path()} timeout : {str(self.timeout)} sec"
            )
        except UnauthorizedClientError as e:
            raise ApiV5AuthenticationError(f"OAUTH2 : Authentication Error : {str(e)}")
        except AccessDeniedError:
            self.access_token, self.refresh_token = None, None
            self.log.warning(
                f"OAUTH2 : (access_denied) invalid token values, reset tokens"
            )
        except Exception as e:
            raise Apiv5ExceptionError(f"OAUTH2 : Error : {str(e)}")
        finally:
            if not self.access_token or self.access_token is None:
                self.log.info(
                    f"OAUTH2 : Access Token for Refresh Token not exist, requests a new Access Token"
                )
                self.get_token()



from typing import Callable

import requests
from requests import Response
from typing_extensions import Literal

from helloasso_api.exceptions import (
    ApiV5BadRequest,
    ApiV5Conflict,
    ApiV5ConnectionError,
    ApiV5Forbidden,
    ApiV5IncorrectMethod,
    ApiV5NoConfig,
    ApiV5NotFound,
    ApiV5RateLimited,
    ApiV5ServerError,
    ApiV5Timeout,
    ApiV5Unauthorized,
)
from helloasso_api.oauth2 import OAuth2Api
from helloasso_api.utils import get_log


class ApiV5Client(object):
    """Manage all calls to Helloasso api (including authentication calls).
    The class must not be used directly but inherited from. See HaApiV5 in src/__init__.py
    """

    def __init__(
        self,
        api_base: str,
        client_id: str,
        client_secret: str,
        timeout: int = None,
        access_token: str = None,
        refresh_token: str = None,
        oauth2_token_getter: Callable[
            [Literal["access_token", "refresh_token"], str], str
        ] = None,
        oauth2_token_setter: Callable[
            [Literal["access_token", "refresh_token"], str, str], None
        ] = None,
    ):
        """
        :param api_base: url of api, example: :api.helloasso-dev.com
        :param client_id: client_id for authentication
        :param client_secret: client_secret for authentication
        :param timeout: (optional) How long to wait for the server to send
            data before giving up, as a float
        :param access_token: (optional) The OAuth access token if exist
        :param refresh_token: (optional) The OAuth refresh token if exist
        :param oauth2_token_getter: custom method to retrieve tokens (useful to share tokens across multiple instances).
        :param oauth2_token_setter: custom method to store tokens (useful to share tokens across multiple instances).
        """
        self.log = get_log("apiv5.apiv5client")

        self.api_base = api_base
        self.timeout = timeout

        self.client_id = client_id
        self.client_secret = client_secret

        if not self.client_id or not self.client_secret:
            raise ApiV5NoConfig("Missing client_id or client_secret.")
        if not self.api_base:
            raise ApiV5NoConfig("Missing Api Base.")

        self.auth = None
        self.access_token = access_token
        self.refresh_token = refresh_token
        self.oauth2_token_getter = oauth2_token_getter
        self.oauth2_token_setter = oauth2_token_setter

        if (oauth2_token_getter is None) != (oauth2_token_setter is None):
            raise ApiV5NoConfig(
                "You must either specify both the oauth2 token setter and getter, or neither."
            )

        self.oauth = OAuth2Api(
            api_base=self.api_base,
            client_id=self.client_id,
            client_secret=self.client_secret,
            timeout=self.timeout,
            access_token=self.access_token,
            refresh_token=self.refresh_token,
            oauth2_token_getter=self.oauth2_token_getter,
            oauth2_token_setter=self.oauth2_token_setter,
        )

        if not self.oauth.access_token:
            self.oauth.get_token()

    def set_access_token(self, access_token: str):
        self.access_token = access_token
        self.oauth.access_token = access_token

    def set_refresh_token(self, refresh_token: str):
        self.refresh_token = refresh_token
        self.oauth.refresh_token = refresh_token

    @staticmethod
    def header() -> dict:
        """Return a base header for all requests."""
        return {"Content-Type": "application/json"}

    def prepare_request(
        self,
        sub_path: str,
        headers: dict,
        data: dict,
        json: dict,
        params: dict,
        include_auth: bool,
    ) -> (str, dict, dict, dict, dict):
        """Build all the elements of the the request."""
        url = f"https://{self.api_base}{sub_path}"
        self.log.debug(f"Prepare Request : {url}")
        data = data or {}
        json = json or {}
        params = params or {}
        headers = headers or {}
        self.auth = (
            {"Authorization": f"Bearer {self.oauth.access_token}"}
            if include_auth
            else {}
        )
        all_headers = {**self.header(), **self.auth, **headers}
        return url, all_headers, data, json, params

    def execute_request(
        self, url: str, method: str, headers: dict, data: dict, json: dict, params: dict
    ) -> Response:
        """Execute request based on method name. Map Api Error to python Exceptions."""
        try:
            self.log.debug(f"Execute Request : {method} : {url}")
            if method == "POST":
                result = requests.post(
                    url,
                    headers=headers,
                    params=params,
                    data=data,
                    json=json,
                    timeout=self.timeout,
                )
            elif method == "GET":
                result = requests.get(
                    url,
                    headers=headers,
                    params=params,
                    data=data,
                    timeout=self.timeout,
                )
            elif method == "PATCH":
                result = requests.patch(
                    url,
                    headers=headers,
                    data=data,
                    timeout=self.timeout,
                )
            elif method == "PUT":
                result = requests.put(
                    url,
                    headers=headers,
                    data=data,
                    timeout=self.timeout,
                )
            elif method == "DELETE":
                result = requests.delete(
                    url,
                    headers=headers,
                    data=data,
                    timeout=self.timeout,
                )
            else:
                raise ApiV5IncorrectMethod(
                    "Incorrect Method: only POST,GET,PATCH,PUT,DELETE authorized."
                )
        except requests.exceptions.Timeout:
            raise ApiV5Timeout(f"{url} timeout : {str(self.timeout)} sec")
        except requests.exceptions.ConnectionError as e:
            raise ApiV5ConnectionError(
                f"Failed to establish a new connection: Name or service not known : {url}"
            )

        if result.status_code in (404, 410):
            raise ApiV5NotFound(result)
        elif result.status_code == 401:
            raise ApiV5Unauthorized(result)
        elif result.status_code == 403:
            raise ApiV5Forbidden(result)
        elif result.status_code == 409:
            raise ApiV5Conflict(result)
        elif result.status_code == 429:
            raise ApiV5RateLimited(result)
        elif 400 <= result.status_code < 500 or result.status_code == 501:
            raise ApiV5BadRequest(result)
        elif result.status_code >= 500:
            raise ApiV5ServerError(result)

        return result

    def call(
        self,
        sub_path: str,
        params: dict = None,
        method: str = "GET",
        data: dict = None,
        json: dict = None,
        headers: dict = None,
        include_auth: bool = True,
    ) -> Response:
        """Manage all api calls. It also handle re-authentication if necessary."""
        self.log.debug(f"Call : {method} : {sub_path}")
        url, headers, data, json, params = self.prepare_request(
            sub_path, headers, data, json, params, include_auth
        )

        try:
            result = self.execute_request(url, method, headers, data, json, params)
            return result
        except ApiV5Unauthorized:
            self.log.warning("401 Unauthorized response to API request.")
            if self.oauth.access_token:
                self.log.info("Refreshing access token")
                self.oauth.refresh_tokens()
            else:
                self.log.info("Get access token")
                self.oauth.get_token()
            return self.call(
                sub_path, params=params, method=method, data=data, headers=headers
            )



from base64 import urlsafe_b64encode
from hashlib import sha256
from random import choices, randint
from urllib.parse import urlencode

from helloasso_api.exceptions import Apiv5ValueError


class AuthorizationApi(object):
    def __init__(self, client):
        self._client = client

    def generate_authorize_request(self, redirect_url: str, state: str = "") -> dict:
        """Generate the needed payload to call the /authorize api endpoint
        :param redirect_url: Helloasso api will redirect to this url after being called.
            This must be the same domain configured in your account for safety reasons.
        :param state: this string will be send back to you in the redirection (max len 500)

        :return dict: {
            "url" : str, ## base url to call to make an authorize call
            "code_verifier": str, ## Will be needed when calling the exchange_authorization_token method
            "params": dict, ## parameters to pass in the authorize call
            "full_url": str, ## A functional GET url to call the /authorize endpoint
        }

        See full documentation at https://api.helloasso-dev.com/v5/swagger/ui/index
        """

        if type(state) is not str or len(state) > 500:
            raise Apiv5ValueError(
                f"state is not valid. Must be a string below 500 characters, "
                f"provided state is {state}"
            )

        chars = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-._~"
        code_verifier = "".join(choices(chars, k=(randint(44, 127))))

        hashed = sha256(code_verifier.encode("ascii")).digest()
        encoded = urlsafe_b64encode(hashed)

        code_challenge = encoded.decode("ascii")[:-1]

        url = self._client.api_base.replace("api", "auth")
        params = {
            "client_id": self._client.client_id,
            "redirect_uri": redirect_url,
            "code_challenge": code_challenge,
            "code_challenge_method": "S256",
            "state": state,
        }

        return {
            "url": f"https://{url}/authorize",
            "code_verifier": code_verifier,
            "params": params,
            "full_url": f"https://{url}/authorize?{urlencode(params)}",
        }

    def exchange_authorization_token(
        self,
        code: str,
        redirect_url: str,
        code_verifier: str,
    ) -> dict:
        """
        Exchange an authorization code with an access_token and a refresh_token
        :param code: The code return by the api callback when calling /authorize endpoint.
        :param redirect_url: The same redirect_url passed to /authorize endpoint.
        :param code_verifier: The string used to generate the code_challenge

        :return dict: {
            "access_token": "token_abc",
            "refresh_token": "token_cba",
            "token_type": "bearer",
            "expires_in": 1800, # seconds
            "organization_slug": "slug_organization"
        }
        """
        headers = {"content-type": "application/x-www-form-urlencoded"}
        body = {
            "client_id": self._client.client_id.replace("-", ""),
            "client_secret": self._client.client_secret,
            "grant_type": "authorization_code",
            "code": code,
            "redirect_uri": redirect_url,
            "code_verifier": code_verifier,
        }
        response = self._client.call(
            "/oauth2/token",
            method="POST",
            data=body,
            headers=headers,
            include_auth=False,
        ).json()

        return {
            "access_token": response["access_token"],
            "refresh_token": response["refresh_token"],
            "token_type": response["token_type"],
            "expires_in": response["expires_in"],  # seconds
            "organization_slug": response["organization_slug"],
        }



from helloasso_api.apiv5client import ApiV5Client
from helloasso_api.client.authorization import AuthorizationApi

"""
Manage all calls to Helloasso api (including authentication calls).
Api endpoints are ordered by domain.

-- Setter and Getter for access/refresh tokens
Refresh and access tokens are stored in apiv5 instance. If you have multiples instances running 
and want to share tokens across instances you can use custom set/get methods.
If you do so you will have to handle storage logic on your side. 
Authentication logic will still be handle by ApiV5Client. 

Example:

def getter(token_key: str) -> str:
    redis.hget("token", token_key)
    
def setter(token_key: str, token: str) -> None:
    redis.hset("token", token_key, token)
    
api = ApiV5(
        client_id="XXXX",
        client_secret="XXXX",
        api_base="XXXX",
        oauth2_token_getter=getter
        oauth2_token_setter=setter
    )
"""


class HaApiV5(ApiV5Client):
    def __init__(self, *args, **kwargs):
        super(HaApiV5, self).__init__(*args, **kwargs)
        self.authorization = AuthorizationApi(self)



class ApiV5Error(ValueError):
    """Any problems get thrown as ApiV5Error exceptions with the relevant info inside"""

    def __init__(self, result, data=None):
        """
        Init ApiV5Error
        :params result : requests result
        :params data: custom data
        """
        super(ApiV5Error, self).__init__(result and getattr(result, "reason", None))

        self.result = result
        self.data = data
        self.error_message = (getattr(result, "reason", "ApiV5Error Error"),)
        self.status_code = (getattr(result, "status_code", ""),)
        self.method = (getattr(result.request, "method", ""),)
        self.url = (getattr(result.request, "url", ""),)
        self.path_url = (getattr(result.request, "path_url", ""),)
        self.headers = (getattr(result.request, "headers", ""),)
        self.body = (getattr(result.request, "body", ""),)
        self.reason = (getattr(result, "content", ""),)
        self.tdata = str(data)

    def __str__(self):
        return str(self.__dict__)


class ApiV5NoConfig(Exception):
    """no client_id or client_secret were passed to the client"""


class ApiV5IncorrectMethod(Exception):
    """ "Incorrect Method: only POST,GET,PATCH,PUT,DELETE authorized."""


class ApiV5Timeout(Exception):
    """socket timeouts, sslerror, and 504"""


class ApiV5AuthenticationError(Exception):
    """Unable to get an token"""


class ApiV5ConnectionError(Exception):
    """Connection Error, example: error with domain"""


class Apiv5ExceptionError(Exception):
    """Generic exception error"""


class Apiv5ValueError(Exception):
    """Value Error"""


class ApiV5BadRequest(ApiV5Error):
    """most 40X results and 501 results"""


class ApiV5NotFound(ApiV5Error):
    """404 and 410 results"""


class ApiV5Unauthorized(ApiV5Error):
    """401 Unauthorized errors"""


class ApiV5Conflict(ApiV5Error):
    """409 conflict errors"""


class ApiV5ServerError(ApiV5Error):
    """most 500 errors"""


class ApiV5RateLimited(ApiV5Error):
    """exception for when we're rate limited"""


class ApiV5Forbidden(ApiV5Error):
    """403 Forbidden"""



from setuptools import setup, find_packages

long_description = '''
Bibliothèque python visant à faciliter l'usage de l'api Helloasso. Compatible de python3.6 jusqu'à python 3.10

La bibliothèque permet de gérér l'authentification et l'authorization auprès de l'api Helloasso

Documentation: https://github.com/HelloAsso/HaApiV5

'''

setup(
    name="helloasso_apiv5",
    version="1.0.0",
    description="Python wrapper for Helloasso APIV5",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/HelloAsso/HaApiV5',
    author="Helloasso",
    author_email="api.support@helloasso.org",
    maintainer='Helloasso',
    license='MIT License',
    packages=find_packages(exclude=("tests", "docs")),
    install_requires=[
        "requests>=2.23.0",
        "oauthlib~=3.1.0",
        "requests-oauthlib~=1.3.0",
        "typing_extensions>=3.7.4.2",
    ],
    python_requires=">=3.6",
)
